package httpsrv

import (
	"encoding/json"
	"io/fs"
	"io/ioutil"
	"net"
	"net/http"
	"strings"
	"sync"

	"git.leolab.info/lib/file"
	"git.leolab.info/lib/logger"
	"git.leolab.info/lib/sess"
	"gorm.io/driver/mysql"
	"gorm.io/driver/postgres"
	"gorm.io/driver/sqlite"
	"gorm.io/gorm"

	"git.leolab.info/lib/errs"
)

//Конфигурация
//	0.1.1
type Config struct {
	DSN  string
	Sess sess.Config
	CORS bool
}

//Основная модель
//	0.1.1
type Server struct {
	Dbg bool
	log *logger.Logger
	ds  *gorm.DB
	cfg *Config

	hande  Handler
	routes *route

	lst net.Listener
	mux *http.ServeMux
	srv *http.Server

	ses *sess.Store

	wg   *sync.WaitGroup
	proc chan interface{}
}

type Handler func(*Rec) *Rec

//Создание экземпляра сервера
//	0.1.0
func New(lst net.Listener, log *logger.Logger) (srv *Server, err *errs.Err) {
	if lst == nil {
		return nil, errs.RaiseError(errs.ErrError, "net.Listener in NIL")
	}
	if log == nil {
		return nil, errs.RaiseError(errs.ErrError, "logger.Logger is NIL")
	}
	srv = &Server{
		log:   log,
		hande: nil,
		lst:   lst,
		mux:   http.NewServeMux(),
		srv:   &http.Server{},
		proc:  make(chan interface{}),
		ses:   nil,
		wg:    &sync.WaitGroup{},
	}
	srv.routes = newRoute(srv)
	return srv, nil
}

//Инициализация сервера
//	0.1.1
func (s *Server) Init(cfg *Config) (err *errs.Err) {
	s.log.Debug("Initializing...")
	s.cfg = cfg
	s.ses, err = sess.NewStore(&cfg.Sess)
	if err != nil {
		return errs.RaiseError(err.Code, err.Msg)
		//return errs.UpError(err)
	}
	s.ds, err = GetDS(cfg.DSN)
	if err != nil {
		return errs.UpError(err)
	}
	s.log.Debug("Initialized")
	return nil
}

//Add static route for path
//	0.1.0
func (s *Server) Static(path string, fs fs.FS) {
	s.log.Debug("Add static route: " + path)
	s.mux.Handle(path, http.FileServer(http.FS(fs)))
}

//Add route
//	0.1.0
func (s *Server) AddRoute(path string, fn Handler) {
	s.log.Debug("Add route: " + path)
	s.routes.add(path, fn)
}

//Default handler, called when no one other hired
//	0.1.0
func (s *Server) DefaultHandle(fn Handler) {
	s.hande = fn
}

//Запуск сервера
//	0.1.0
func (s *Server) Start() {
	s.log.Info("Starting...")
	s.mux.HandleFunc("/", s.reqHandler)
	s.wg.Add(1)
	go s.exec()
}

//Останов сервера
//	0.1.0
func (s *Server) Stop() {
	s.log.Info("Stopping...")
	close(s.proc)
	s.wg.Wait()
	s.log.Info("Stopped")
}

//Выполнение сервера
//	0.1.0
func (s *Server) exec() {
	defer s.wg.Done()
	s.srv.Handler = s.mux
	go func() {
		if e := s.srv.Serve(s.lst); e != nil {
			err := errs.RaiseError(ErrServerError, e.Error())
			s.log.Error(err.Error(), map[string]interface{}{"Error": err})
		}
	}()
	s.log.Info("Started")
	<-s.proc
	s.log.Info("Finishing...")
}

//Основной обработчик
//	0.1.1
func (s *Server) reqHandler(w http.ResponseWriter, r *http.Request) {
	rl := rlStart(s)
	/*
		defer func(rl *ReqRec) {
			if r := recover(); r != nil {
				er := errs.RaiseError(ErrPanic, "Panic while proceed '"+rl.ReqUrl+"' request")
				s.log.Error(er.Error(), map[string]interface{}{"Panic": r})
				rl.Finish(er)
			} else {
				rl.Finish(nil)
			}
		}(rl)
	*/
	if r.URL.Path[len(r.URL.Path)-1] == '/' {
		r.URL.Path = r.URL.Path + "index"
	}
	rl.ReqUrl = r.URL.Path
	fa := strings.Split(strings.TrimLeft(r.URL.Path, "/"), "/")
	if len(fa) == 0 {
		fa = append(fa, "index")
	}

	rl.Addr = r.RemoteAddr
	rl.XAddr = r.Header.Get("X-Forwarded-For")
	rl.UserAgent = r.Header.Get("User-Agent")

	b, e := ioutil.ReadAll(r.Body)
	if e != nil {
		er := errs.RaiseError(ErrReadBodyError, e.Error()+" for: "+r.URL.Path)
		s.log.Error(er.Error())
		b = make([]byte, 0)
	}
	sr := &Rec{
		Cors: s.cfg.CORS,
		Req: &Request{
			R:    r,
			Body: b,
			F:    fa[0],
			FA:   fa[1:],
			Lnk:  strings.Join(fa, "/"),
			Vars: make(map[string]string),
		},
		Rsp: &Response{
			W:       w,
			headers: make(map[string][]string),
			code:    200,
		},
		Data: make(map[string]interface{}),
		//Sess: s._ses.Get(w, r),
		_content: make([]byte, 0),
		_sent:    false,
	}

	var err *errs.Err
	sr.Sess, err = s.ses.Get(w, r)
	if err != nil {
		er := errs.RaiseError(err.Code, "Session ERROR: "+err.Error(), err)
		s.log.Error("Session ERROR: " + err.Error())
		rsp := Error(sr, er.Error())
		rsp.Sent()
		rl.Finish(er)
		return
	}
	rl.SSID = sr.Sess.SSID

	//s.log.Debug("Request: " + sr.Req.Lnk)

	if e := json.Unmarshal(b, &sr.Req.Data); e != nil {
		//No JSON data in body, or wrong
		s.log.Debug("No JSON data for: " + sr.Req.Lnk)
		sr.Req.Data = nil
	} else {
		s.log.Debug("JSON Data for "+sr.Req.Lnk, sr.Req.Data)
	}
	rl.ReqData = sr.Req.Data
	q := r.URL.Query()
	for k, v := range q {
		sr.Req.Vars[k] = strings.Join(v, ",")
	}
	for k, v := range sr.Req.Vars {
		rl.ReqVars[k] = v
	}

	rsp := s.routes.Do(sr)
	if rsp == nil {
		er := errs.RaiseError(ErrServerError, "NIL response")
		s.log.Error("NIL response: " + sr.Req.Lnk)
		w.WriteHeader(500)
		w.Write([]byte(er.Error()))
		rl.Finish(er)
		return
	}
	if !rsp._sent {
		rsp.Sent()
	}
	rl.Finish(nil)
}

func GetDS(dsn string) (ds *gorm.DB, err *errs.Err) {
	if dsn == "" {
		return nil, nil
	}
	var e error
	cfg := &gorm.Config{}
	dsa := strings.Split(dsn, ">")
	if len(dsa) < 2 {
		return nil, errs.RaiseError(ErrConfigError, "Invalid DSN: "+dsn)
	}
	switch dsa[0] {
	case "sqlite":
		ds, e = gorm.Open(sqlite.Open(file.RealPath(dsa[1])), cfg)
		if e != nil {
			return nil, errs.RaiseError(errs.ErrDSError, "'"+dsa[1]+"'"+e.Error())
		}
	case "mysql":
		ds, e = gorm.Open(mysql.Open(dsa[1]), cfg)
		if e != nil {
			return nil, errs.RaiseError(errs.ErrDSError, e.Error())
		}
	case "postgres":
		ds, e = gorm.Open(postgres.New(postgres.Config{
			DSN: dsa[1],
		}), cfg)
		if e != nil {
			return nil, errs.RaiseError(errs.ErrDSError, e.Error())
		}
	default:
		return nil, errs.RaiseError(errs.ErrDSError, "Incorrect DSN type: "+dsa[0])
	}
	if e := ds.AutoMigrate(&ReqRec{}); e != nil {
		return nil, errs.RaiseError(errs.ErrDSError, e.Error())
	}
	return ds, nil
}
